def wrap_purple(string):
  return f'\033[95m{string}\033[0m'

def wrap_blue(string):
  return f'\033[94m{string}\033[0m'

def wrap_green(string):
  return f'\033[92m{string}\033[0m'

def wrap_red(string):
  return f'\033[91m{string}\033[0m'

def wrap_orange(string):
  return f'\033[93m{string}\033[0m'
  
def bold(string):
  return f'\033[1m{string}\033[0m'

def underline(string):
  return f'\033[4m{string}\033[0m'


import argparse, sys, re, os
import dateutil.parser as dateparser
from .utility_functions import *
from .connection_tools import *
from .color_wrappers import *
from dateutil import tz
import pytz

def fetch_arguments():
  ap = argparse.ArgumentParser()
  ap.add_argument("-aep", "--aggregated_export_path", required=False, help="path where to store the aggregated csv file. If not set, 'aggregated_es_export.csv' will be used", type=check_csv_valid_filename, default='aggregated_es_export.csv')
  ap.add_argument("-af", "--aggregation_fields", required=False, help="set this option if you want to generate an additional file (raw exports file will still be generated) containing the info aggregated according to specific fields. Specify the fields to aggregate on as a string with commas between fields and no whitespaces (e.g. \"field1,field2\")", default=None)
  ap.add_argument("-ao", "--aggregate_only", required=False, help="set to True to skip all the extraction and processing part. Default to False. In case you enable this option you have to pass an input csv raw file (-rif/--raw_input_file) to aggregate on or an exception will be raised.", type=real_bool, default=False)
  ap.add_argument("-asi", "--allow_short_interval", required=False, help="set this option to True to allow the --load_balance_interval to go below 1 day. With this option enabled the --load_balance_interval can be set up to 1 minute (1m)", default=False)
  ap.add_argument("-at", "--aggregation_types", required=False, help="fields to aggregate and relative aggregation functions to use when generating the aggregated csv file. Specify those fields who require a different type of aggregation. It can be one of the following: ['count', 'min', 'max', 'mean', 'sum']. You can set them as a string with each field to aggregate linked to the following aggregation type through a double dash (--) and separated from the next couple of values with a comma, like \"agg_var1--count,agg_var2--sum\". Default value is your last aggregation field and the count aggregation, like \"agg_field3--count\". Requires the --aggregation_fields to be set.", type=check_valid_dictionary_couples, default=None)
  ap.add_argument("-atf", "--aggregation_time_field", required=False, help="in case you want to aggregate according to a specific time field, specify here which field you want to consider. Obviously it must be among the fields extracted from the index or among those in csv file imported with the (-ao + -rif) combination", default=None)
  ap.add_argument("-ats", "--aggregation_time_span", required=False, help="number of days you want to aggregate the -atf on. Default to one day (1)", type=int, default=1)
  ap.add_argument("-b", "--batch_size", required=False, help="batch size for the scroll API. Default to 5000. Max 10000. Increasing it might impact the ES instance heap memory. If you want to set a value greater than 10000, you must set the max_result_window elasticsearch property accordingly first. Please check out the elasticsearch documentation before increasing that value on the specified index --> https://www.elastic.co/guide/en/elasticsearch/reference/current/index-modules.html", type=int, default=5000)
  ap.add_argument("-c", "--cert_verification", required=False, help="require ssl certificate verification. Default to False. Set to True to enable it", type=real_bool, default=False)
  ap.add_argument("-cbo", "--count_boolean_occurrences", required=False, help="this option might help you in case you want to count the occurrences of a boolean field. All boolean fields that have to be taken into consideration when True, must be passed as a string with commas between fields and no whitespaces (e.g. \"field1,field2\"). Remember to set the aggregation type \"sum\" in the --aggregation_types option for the corresponding aggregation fields", default='')
  ap.add_argument("-cp", "--certificate_path", required=False, help="path to the certificate to verify the instance certificate against", default='')
  ap.add_argument("-cs", "--csv_separator", required=False, help="separator to use in the CSV resulting files (both raw and aggregated, if any). Default is comma (,). Separators longer than 1 character and different from '\s+' will be interpreted as regular expressions and will also force the use of the Python parsing engine. Note that regex delimiters are prone to ignoring quoted data", default=',')
  ap.add_argument("-dp", "--disable_progressbar", required=False, help="turn off the progressbar visualization useful to keep track of fetching data progresses of various processes. Default to False. Set to True to simply be noticed when processes have done fetching data, without the loading progressbar.", type=real_bool, default=False)
  ap.add_argument("-dr", "--decimal_rounding", required=False, help="round decimals to this value. Default to 2", type=int, default=2)
  ap.add_argument("-ds", "--decimal_separator", required=False, help="set the custom decimal separator for your file. Consider one that doesn't mess up with the csv_separator (default to ','). Default is the dot (.)", default='.')
  ap.add_argument("-e", "--export_path", required=False, help="path where to store the csv file. If not set, 'es_export.csv' will be used. Make sure the user who's launching the script is allowed to write to that path. WARNING: At the end of the process, unless --keep_partial is set to True, all the files with filenames \"[--export_path]_process*.csv\" will be remove. Make sure you're setting a --export_path which won't accidentally delete any other file apart from the ones created by this script", type=check_csv_valid_filename, default="es_export.csv")
  ap.add_argument("-ed", "--ending_date", required=False, help="query ending date. Must be set in iso 8601 format, without the timezone that can be specified in the --timezone option (e.g. \"YYYY-MM-ddTHH:mm:ss\")", default='now+1000y')
  ap.add_argument("-em", "--enable_multiprocessing", required=False, help="enable the multiprocess options. Default to False. Set to True to exploit multiprocessing. If set to True a --time_field to sort on must be set or an exception will be raised", type=real_bool, default=False)
  ap.add_argument("-f", "--fields", required=True, help="Elasticsearch fields, passed as a string with commas between fields and no whitespaces (e.g. \"field1,field2\")")
  ap.add_argument("-ho", "--host", required=False, help="Elasticsearch host. If not set, localhost will be used", default="localhost")
  ap.add_argument("-i", "--index", required=True, help="Elasticsearch index pattern to query on. To use wildcard (*) put the index in quotes (e.g. \"my-indices*\")")
  ap.add_argument("-k", "--keep_partials", required=False, help="during the processing, various partial csv files will be created before joining them into a single csv. Set this flas to True if you want to keep also these partial files. Default to False. Notice the partial files will be kept anyway if something goes wrong during the creation of the final file.", type=real_bool, default=False)
  ap.add_argument("-lbi", "--load_balance_interval", required=False, help="set this option to build process intervals by events count rather than equally spaced over time. The shorter the interval, the better the events-to-process division, the higher the heavier the computation. It cannot go below 1d if --allow_short_interval is not set. Allowed values are a number (integer only) plus one of the following [m, h, d, w, M, y], like 1d for 1 day or 4M for 4 months. Multiprocessing must be enabled to set this option", default=None)
  ap.add_argument("-mf", "--metadata_fields", required=False, help="Elasticsearch metadata fields (_index, _type, _id, _score), passed as a string with commas between fields and no whitespaces (e.g. \"_id,_index\")", default='')
  ap.add_argument("-o", "--scroll_timeout", required=False, help="scroll window timeout. Default to 4m", default='4m')
  ap.add_argument("-p", "--port", required=False, help="Elasticsearch port. If not set, the default port 9200 will be used", default=9200, type=int)
  ap.add_argument("-pcs", "--partial_csv_size", required=False, help="Max number of rows of partial csv files created by a single process. Default to 10000000. The higher the number of fields to extract, the lower this number should be not to keep too many data in memory. If set, must be greater than --batch_size (default 5000)", type=int, default=10000000)
  ap.add_argument("-pn", "--process_number", required=False, help="number of processes to run the script on. Default to max number of processes for the hosting machine", type=int)
  ap.add_argument("-pw", "--password", required=False, help="Elasticsearch password in clear. If set, the --secret_password will be ignored. If both this a --secret_password are not set, a prompt password will be asked anyway (leave it blank if not needed).", default=None)
  ap.add_argument("-q", "--query_string", required=False, help="Elasticsearch query string. Put it between quotes and escape internal quotes characters (e.g. \"one_field: foo AND another_field.keyword: \\\"bar\\\"\"", default="*")
  ap.add_argument("-ra", "--rename_aggregations", required=False, help="if you enable any aggregation types on a field, default name for the resulting column will be \"estocsv__\{field_name\}__\{aggregation_type\}\", like \"estocsv_myfield_count\". If you want to set custom names for the resulting columns, you can specify them here, as a string with each field to aggregate linked to the following aggregation name through a double dash (--) and separated from the next couple of values with a comma, like \"field1--count_field1,field2--sum_field2\".. Requires --aggregation_fields to be set", type=check_valid_dictionary_couples, default=None)
  ap.add_argument("-rd", "--remove_duplicates", required=False, help="set to True to remove all duplicated events. Default to False. WARNING: two events with the same values of the fields specified in --fields will be considered duplicated and then unified even if on ES they might not be equal because of other fields not included in --fields. Check out the --metadata_fields option to include further info like the ES _id", default=False)
  ap.add_argument("-rf", "--rename_fields", required=False, help="you can set this option to rename some of your fields right before the csv writing. You can set them as a string with each field to rename linked to the following new name through a double dash (--) and separated from the next couple of values with a comma, like \"agg_var1--count,agg_var2--sum\". Default value is your last aggregation field and the count aggregation, like \"agg_field3--count\". WARNING: All the manipulation and aggregations, if any, will impact the original name of the fields so do not use these new names in other arguments but stick to the original ones.", type=check_valid_dictionary_couples, default={})
  ap.add_argument("-rif", "--raw_input_file", required=False, help="required if --aggregate_only is enabled. Path to csv file to import as raw file", default=None)
  ap.add_argument("-s", "--ssl", required=False, help="require ssl connection. Default to False. Set to True to enable it", type=real_bool, default=False)
  ap.add_argument("-sd", "--starting_date", required=False, help="query starting date. Must be set in iso 8601 format, without the timezone that can be specified in the --timezone option (e.g. \"YYYY-MM-ddTHH:mm:ss\")", default='now-1000y')
  ap.add_argument("-sm", "--silent_mode", required=False, help="set this option to True to run the tool in silent mode, without any user interaction (password request if not specified in -pw or -spw, warning about files overwriting). If you enable this option, consider you won't be asked for anything, so an error will be returned in case a password is missing and files might be overwritten without any confirmation first. Default to False.", type=real_bool, default=False)
  ap.add_argument("-spw", "--secret_password", required=False, help="env var pointing the Elasticsearch password. If both this a --password are not set, a prompt password will be asked anyway (leave it blank if not needed).", default=None)
  ap.add_argument("-t", "--time_field", required=False, help="time field to query on. If not set and --starting_date or --ending_date are set and exception will be raised", default=None)
  ap.add_argument("-tz", "--timezone", required=False, help="timezone to set according to the time zones naming convention (e.g. \"America/New_York\" or \"Europe/Paris\" or \"UTC\"). If not set, the local timezone of the present machine will be used", default=None)
  ap.add_argument("-u", "--user", required=False, help="Elasticsearch user", default='')
  return vars(ap.parse_args())

def check_csv_valid_filename(filename):
  if filename[-4:] != '.csv':
    raise argparse.ArgumentTypeError(wrap_red(f"{filename} is not a valid option. It must be a .csv file"))
  return filename

def check_valid_dictionary_couples(string):
  try:
    [couple.split('--') for couple in string.split(',')]
    return string if string else {}
  except:
    sys.exit(wrap_red(f"Something's wrong with the value you inserted {string}. Remember this value must be a string of couples separated by commas. Each couple must have its value separated by two dashes. Like \"key1--value1,key2--value2\""))

def real_bool(stringified_bool):
  if stringified_bool.lower().strip() not in ['true', 'false']: raise argparse.ArgumentTypeError(wrap_red(f"{stringified_bool} must be a bool type. Please insert 'true' or 'false'"))
  return True if stringified_bool.lower().strip() == 'true' else False

def check_valid_date(date_string):
  if (date_string != "now-1000y" and date_string != "now+1000y"):
    try:
      dateparser.parse(date_string)
    except:
      sys.exit(wrap_red(f"\nThe date set ({date_string}) is not valid"))
  return True

def aggregation_log(filename, aggregation_fields, aggregation_types):
  if aggregation_fields is None:
    return wrap_blue("No --aggregation_fields value has been set. Aggregated csv file won't be created\n")
  else:
    return wrap_blue(f"A file \"{filename}\" will be created according to the value of --aggregation_fields ({aggregation_fields}) and --aggregation_types ({aggregation_types})\n")

def valid_bound_dates(args):
  if args['starting_date'] != 'now-1000y' and args['ending_date'] != 'now+1000y':
    sdate = dateparser.parse(args['starting_date']).astimezone(args['timezone'])
    edate = dateparser.parse(args['ending_date']).astimezone(args['timezone'])
    return edate > sdate
  return True

def check_timezone_validity(timezone, log):
  if timezone is None:
    return tz.tzlocal()
  elif timezone in pytz.all_timezones:
    return tz.gettz(timezone)
  else:
    log.error(wrap_red(f"\n\nSomething is wrong with the timezone you set {timezone}. Please set a timezone included in the pytz.all_timezones or leave it blank to set the local timezone of this machine"))
    os._exit(os.EX_OK)

def check_meta_fields(meta_fields_str):
  try:
    if meta_fields_str == '': return []

    meta_fields = meta_fields_str.split(',')
    for mf in meta_fields:
      if mf not in ['_index', '_type', '_id', '_score']:
        sys.exit(wrap_red(f"One of your --metadata_fields {mf} is not allowed. Allowed metadata fields are [_index, _type, _doc, _score]. Check out the --help to know how to set them."))
    return meta_fields
  except Exception as e:
    sys.exit(wrap_red(f"Something is wrong with the --metadata_fields you set or how you set them. Check out the --help to know how to set them. Here's the exception:\n\n{e}"))

def check_fields(fields_str):
  try:
    fields = fields_str.split(',')
    return fields
  except Exception as e:
    sys.exit(wrap_red(f"Something is wrong with the --fields you set. Check out the --help to know how to set them. Here's the exception:\n\n{e}"))

def parse_lbi(lbi, allow_short_interval, multiprocess_enabled):
  try:
    number = re.search("\d+", lbi).group() if re.search("\d+", lbi) != None else None
    unit = re.search("[^\d]+", lbi).group() if re.search("[^\d]+", lbi) != None else None
    allowed_units = ['m', 'h', 'd', 'w', 'M', 'y'] if allow_short_interval else ['d', 'w', 'M', 'y']
    unit_in_seconds = {'m':60, 'h':3600, 'd':86400, 'w':604800 , 'M':2592000, 'y':31104000}
    if number == None or not number.isnumeric():
      sys.exit(wrap_red("--load_balance_interval option must begin with a number. Please check the --help to know how to properly set it"))
    elif unit == None or unit == '' or unit not in allowed_units:
      sys.exit(wrap_red(f"--load_balance_interval unit must be one of the following {allowed_units}. Please check the --help to know how to properly set it"))
    elif not multiprocess_enabled:
      sys.exit(wrap_red("Multiprocessing must be enabled (-em True) in order to set the --load_balance_interval"))
    else:
      return int(number) * unit_in_seconds[unit]
  except:
    sys.exit(wrap_red("Something in the combination of --load_balance_interval and --allow_short_interval you set is wrong. Please check the --help to know how to properly set them"))

def check_valid_lbi(starting_date, ending_date, lbi):
  sdate_in_seconds = dateparser.parse(starting_date).timestamp()
  edate_in_seconds = dateparser.parse(ending_date).timestamp()
  if lbi >= (edate_in_seconds - sdate_in_seconds):
    sys.exit(wrap_red("You set a --load_balance_interval greater than the timespan [--starting_date - --ending_date]. You might as well avoid the multiprocessing :)"))
  return True

def check_valid_aggregations(aggregation_types, aggregation_fields, log):
  valid_aggregation_types = ['sum', 'count', 'max', 'min', 'mean']
  if not aggregation_types:
    log.info(wrap_blue(f"No aggregeation type was specified. Count aggregation will be made on {aggregation_fields[-1]} field\n"))
    return { aggregation_fields[-1]: 'count' }
  else:
    aggregations = {}
    agg_couples = [couple.split('--') for couple in aggregation_types.split(',')]
    for couple in agg_couples:
      if len(couple) == 1:
        log.info(wrap_blue(f"No aggregeation type was specified for field {couple[0]}. Count aggregation will be made on this field\n"))
        aggregations[couple[0]] = 'count'
      elif len(couple) == 2 and couple[1] in valid_aggregation_types:
        aggregations[couple[0]] = couple[1]
      else:
        sys.exit(wrap_red(f"\n\nSomething's wrong with the --aggregation_types you set. If you set this field, it must be a string like \"field1--agg_type1,field2--agg_type2\". If you exported the raw data from Elasticsearch, the raw csv file will be available at the export_path. Please use it with the --aggregate_only flag and a proper --aggregation_types and --aggregation_fields configuration to generate your aggregated file."))
    return aggregations


  if aggregation_type.strip().lower() not in ['sum', 'count', 'max', 'min', 'mean']: sys.exit(wrap_red(f"Aggregation function you set ({aggregation_type}) is not a valid one. It must be one of ['count', 'min', 'max', 'mean', 'sum']"))
  return aggregation_type.strip().lower()

def check_valid_rename(renamed_aggregations, log):
  if not renamed_aggregations:
    return {}
  else:
    aggregation_names = {}
    name_couples = [couple.split('--') for couple in renamed_aggregations.split(',')]
    for couple in name_couples:
      if len(couple) == 2:
        aggregation_names[couple[0]] = couple[1]
      else: 
        log.info(wrap_orange(f"Something's wrong with the renaming of the aggregations. \"{couple}\" is not a valid value. Default one will be used\n"))
    return aggregation_names

def get_actual_bound_dates(args, starting_date, ending_date):
  search_url = "{url_prefix}://{host}:{port}/{index}/_search".format(**args)
  timezone = args['timezone']
  starting_date = add_timezone(starting_date, timezone) if not starting_date == "now-1000y" else starting_date
  ending_date = add_timezone(ending_date, timezone) if not ending_date == "now+1000y" else ending_date
  # Fetch date of first element from the specified starting_date
  sdate_query = build_es_query(args, starting_date, ending_date, 'asc', 1, source=args['time_field'].split())
  r = request_to_es(search_url, sdate_query, args['log'], args['user'], args['password'], verification=args['certificate_path'])
  starting_date = add_timezone(r['hits']['hits'][0]['_source'][args['time_field']], timezone)
  # Fetch date of last element before the specified ending_date
  edate_query = build_es_query(args, starting_date, ending_date, 'desc', 1, source=args['time_field'].split())
  r = request_to_es(search_url, edate_query, args['log'], args['user'], args['password'], verification=args['certificate_path'])
  ending_date = add_timezone(r['hits']['hits'][0]['_source'][args['time_field']], timezone)
  # Return real starting_date and ending_date with proper timezone
  return [starting_date, ending_date]

def file_not_found(filename):
  if filename is None: sys.exit(wrap_red(f"If -ao/--aggregate_only is set, a -rif/--raw_input_file to fetch the data from must be set, too."))
  if os.path.exists(filename):
    check_csv_valid_filename(filename)
    return False
  else: 
    return True

def check_valid_agg_related_fields(args, log):
  if args['aggregate_only'] and file_not_found(args['raw_input_file']): sys.exit(wrap_red("\nIf you want to simply make aggregations (-ao/--aggregate_only) on an already existing file, you have to pass the -rif/--raw_input_file to make aggregations on, don't you think? :-) "))
  if args['aggregation_fields'] is None: sys.exit(wrap_red("\n-af/--aggregation_fields is required to perform any aggregation"))
  args['aggregation_types'] = check_valid_aggregations(args['aggregation_types'], args['aggregation_fields'].split(','), log)
  args['rename_aggregations'] = check_valid_rename(args['rename_aggregations'], log)
  args['aggregation_fields_checked'] = True
  return args

def check_arguments_conflicts(args, log):
  args['log'] = log
  args['aggregation_fields_checked'] = False
  if args['rename_fields']: args['rename_fields'] = dict([couple.split('--') for couple in args['rename_fields'].split(',') if len(couple.split('--')) == 2])

  if args['aggregate_only']:
    args = check_valid_agg_related_fields(args, log)
    args['df_ready'] = None
    return args

  if (args['starting_date'] != 'now-1000y' or args['ending_date'] != 'now+1000y') and args['time_field'] == None:
    sys.exit(wrap_red("\nIf you set either a starting_date or an ending_date you have to set a --time_field to sort on, too."))
  
  args['verify'] = False if args['cert_verification'] is False else True if args['certificate_path'] == '' else args['certificate_path']

  if args['enable_multiprocessing'] and args['time_field'] == None:
    sys.exit(wrap_red("\nYou have to set a --time_field in order to use multiprocessing."))

  if args['batch_size'] >= args['partial_csv_size']: sys.exit(wrap_red(f"\n--partial_csv_size ({args['partial_csv_size']}) must be greater than --batch_size ({args['batch_size']})"))

  args['password'] = final_pw(args, log)
  
  args['url_prefix'] = 'https' if args['ssl'] else 'http'

  test_es_connection(args, log)
  args['count_url'] = "{url_prefix}://{host}:{port}/{index}/_count".format(**args)

  args['timezone'] = check_timezone_validity(args['timezone'], log)
  check_valid_date(args['starting_date'])
  check_valid_date(args['ending_date'])
  if not valid_bound_dates(args):
    sys.exit(wrap_red(f"\nThe --starting_date you set ({args['starting_date']}) comes after the --ending_date ({args['ending_date']}). Please set a valid time interval"))
  if args['time_field'] != None: args['starting_date'], args['ending_date'] = get_actual_bound_dates(args, args['starting_date'], args['ending_date']) 

  args['fields'] = check_fields(args['fields'])

  args['metadata_fields'] = check_meta_fields(args['metadata_fields'])
  args['fields_to_export'] = args['metadata_fields'] + args['fields']

  if args['load_balance_interval'] != None: 
    args['load_balance_interval'] = parse_lbi(args['load_balance_interval'], args['allow_short_interval'], args['enable_multiprocessing']) 
    check_valid_lbi(args['starting_date'], args['ending_date'], args['load_balance_interval'])


  return args


from elasticsearch import Elasticsearch
from elasticsearch import RequestsHttpConnection
from requests.exceptions import SSLError as RequestsSSLError
import requests, sys, json
from ssl import create_default_context
from requests.auth import HTTPBasicAuth
from .utility_functions import *
from tqdm import *
from .csv_handlers import *
from .color_wrappers import *

def build_es_connection(args):
  return Elasticsearch( hosts=[{'host': args['host'], 'port': args['port']}],
                          connection_class=RequestsHttpConnection,
                          http_auth=(args['user'], args['password']),
                          scheme=args['url_prefix'],
                          verify_certs=args['verify'],
                          retry_on_timeout=True,
                          timeout=50, ssl_show_warn=False )


def request_to_es(url, query, log, user='', pwd='', timeout=10, verification=False):
  headers = {'content-type': 'application/json', 'Accept-Charset': 'UTF-8'}
  try:
    r = requests.get(url, data=query, headers=headers, auth=HTTPBasicAuth(user, pwd), timeout=10, verify=verification).json()
  except Exception as e:
    log.error(wrap_red(f"\n\nSomething when wrong connecting to the ES instance. Check out the raised exception: \n\n{e}"))
    os._exit(os.EX_OK)
  return r

def test_es_connection(args, log):
  try:
    url = "{url_prefix}://{host}:{port}".format(**args)
    headers = {'content-type': 'application/json', 'Accept-Charset': 'UTF-8'}
    check_valid_certificate_combinations(args, log)
    r = requests.get(url, headers=headers, auth=HTTPBasicAuth(args['user'], args['password']), timeout=10, verify=args['verify'])
    if r.status_code != 200: sys.exit(wrap_red(f"Status code when trying to connect to your host at {url} is not 200. Check out the reason here:\n\n{json.dumps(r.json(), indent=2)}"))
  except Exception as e:
    sys.exit(wrap_red(f"Something went wrong when testing the connection to your host. Check your host, port, credentials and certificates (if not ignored). Check your ES instance is still running, too. Here's the exception:\n\n{e}"))

def check_valid_certificate_combinations(args, log):
  if args['cert_verification'] is True and args['certificate_path'] == '':
    log.warning(wrap_orange(f"You set --cert_verification to True but no --certificate_path to fetch a local certificate. If the certificate of the ES instance you are trying to connect to doesn't have a root CA in its CA chain (e.g. self-signed certificate) the validation is going to fail.\n"))
    test_certificate(args, "You didn't set a path to a certificate yet you required a certificate verification. Are you sure your ES instance has a certificate signed by a valid root CA? ")
  elif args['cert_verification'] is True:
    test_certificate(args, f"There's a mismatch between the certificate you passed ({args['certificate_path']}) and the one of the ES instance. ")
  elif args['ssl'] is True and args['cert_verification'] is False:
    log.warning(wrap_orange("You're connecting to a ES instance over SSL but without any certificate verification. Be sure you know the risks (e.g. MITM attack)."))
    proceed_without_cert_verification = 'y' if args['silent_mode'] else ''
    while proceed_without_cert_verification not in ['y', 'n']:
      proceed_without_cert_verification = input(wrap_orange(f"\nDo you want to proceed? (y/n)")).lower().strip()
      if proceed_without_cert_verification == 'n':
        sys.exit(wrap_red(wrap_blue("\nExiting script. No connection to the host has been established.")))
      elif proceed_without_cert_verification != 'y':
        print("\nSorry, I can't understand the answer. Please answer with 'y' or 'n'")

def test_certificate(args, exception=''):
  try: 
    url = "{url_prefix}://{host}:{port}".format(**args)
    headers = {'content-type': 'application/json', 'Accept-Charset': 'UTF-8'}
    r = requests.get(url, headers=headers, auth=HTTPBasicAuth(args['user'], args['password']), timeout=10, verify=args['verify'])
  except RequestsSSLError as e:
    sys.exit(wrap_red(f"Something's wrong with the certificate validation. {exception}\nIf you don't have a valid certificate to pass to --certificate_path, you can work around the problem ignoring the certificate verification (-c False) but check out MITM attack risks first. Here's the exception:\n\n{e}"))

def fetch_es_data(args, starting_date, ending_date, process_name='Main'):
  log = args['log']
  process_number = 0 if process_name == 'Main' else process_name
  process_tmp_subset = 1
  csv_partial_filename = f"{args['export_path'][:-4]}_process{process_number}_{str(process_tmp_subset).zfill(5)}.csv"
  log.info(f"Process {process_name}: starts fetching data from {starting_date} to {ending_date}")
  df_header = ['_id'] + args['fields_to_export'] if not '_id' in args['fields_to_export'] else args['fields_to_export']
  meta_for_extraction = ['_id'] + args['metadata_fields'] if not '_id' in args['metadata_fields'] else args['metadata_fields']
  es_count_query = build_es_query(args, starting_date, ending_date, count_query=True)

  total_hits = request_to_es(args['count_url'], es_count_query, log, args['user'], args['password'], verification=args['certificate_path'])['count']
  pbar = tqdm(total=total_hits, position=process_number, leave=False, desc=f"Process {process_name} - Fetching", ncols=150, mininterval=0.05) if not args['disable_progressbar'] else None
  
  fetched_data = []

  es_instance = build_es_connection(args)
  es_index = args['index']
  scroll_timeout = args['scroll_timeout']
  batch_size = args['batch_size']
  es_query = build_es_query(args, starting_date, ending_date, source=args['fields'])
  try:
    es_data = es_instance.search(
      index = es_index,
      _source = args['fields'],
      scroll = scroll_timeout,
      size = batch_size,
      body = es_query
    )
  except Exception as e:
    log.error(wrap_red(f"Process {process_name}: something went wrong when fetching the data from Elasticsearch. Please check your connection parameters. Here's the raised exception: \n\n{e}"))
    os._exit(os.EX_OK)
  
  # Save parameters for scrolling
  sid = es_data['_scroll_id']
  scroll_size = len(es_data['hits']['hits'])

  # Process current batch of hits before starting to scroll
  for hit in es_data['hits']['hits']:
    fetched_data.append(denest(args['fields'], add_meta_fields(hit, meta_for_extraction, log))) 

  # Scroll and add hits to the fetched_data list
  while scroll_size > 0:
    es_data = es_instance.scroll(scroll_id=sid, scroll=scroll_timeout)
    # Process current batch of hits
    for hit in es_data['hits']['hits']:
      fetched_data.append(denest(args['fields'], add_meta_fields(hit, meta_for_extraction, log))) 
      if not args['disable_progressbar']: pbar.update(1)
    # Update the scroll ID
    sid = es_data['_scroll_id']
    # Get the number of results that returned in the last scroll
    scroll_size = len(es_data['hits']['hits'])

    # If this process has already fetched 10M events, create the df, write the partial csv and empty the fetched_data list
    if len(fetched_data) >= args['partial_csv_size']:
        write_csv(csv_partial_filename, df_header,  args['csv_separator'], args['decimal_separator'], args['decimal_rounding'], exception_message=f"Something went wrong when trying to write the partial csv {csv_partial_filename}.", list_to_convert=fetched_data)
        process_tmp_subset += 1
        csv_partial_filename = f"{args['export_path'][:-4]}_process{process_number}_{str(process_tmp_subset).zfill(5)}.csv"
        fetched_data = []
    
  write_csv(csv_partial_filename, df_header,  args['csv_separator'], args['decimal_separator'], args['decimal_rounding'], exception_message=f"Something went wrong when trying to write the partial csv {csv_partial_filename}.", list_to_convert=fetched_data)
  log.info(bold(f"Process {process_name} has fetched and processed {total_hits} docs. They've been split into {process_tmp_subset} partial csv file(s)"))
  return True



import os, logging, getpass, sys
import dateutil.parser as dateparser
import pandas as pd
import numpy as np
from dateutil import tz
from .color_wrappers import *

def build_source_query(fields_of_interest):
  return  '\"_source\":' + str(fields_of_interest).replace("'",'"') + ','

def build_es_query(args, starting_date, ending_date, order='asc', size=None, count_query=False, source=[]):
  QUERY_STRING = args['query_string'] #TODO chech how to properly escape internal quotes
  SIZE = ('"size": ' + str(size) + ',') if size != None and not count_query else ''
  if args['time_field'] != None:
    TIME_FIELD = args['time_field']
    SORT_QUERY = '"sort":[{"' + TIME_FIELD + '":{"order":"' + order + '"}}],' if not count_query else ''
    RANGE_QUERY = ",{\"range\":{\"" + TIME_FIELD + "\":{\"gte\":\"" + starting_date + "\",\"lte\":\"" + ending_date + "\"}}}"
  else:
    SORT_QUERY = RANGE_QUERY = ''
  SOURCE_QUERY = build_source_query(source) if not source == [] and not count_query else ''
  ES_QUERY = '{' + SOURCE_QUERY + SIZE + SORT_QUERY + '"query":{"bool":{"must":[{"query_string":{"query":"' + QUERY_STRING + '"}}' + RANGE_QUERY + ']}}}'
  return ES_QUERY

def add_meta_fields(obj, meta_fields, log=logging):
  try:
    for index, mf in enumerate(meta_fields):
      obj['_source'][mf] = obj[mf]
    return obj['_source']
  except Exception as e:
    log.critical(wrap_red(f"Something is wrong with the metadata retrieval in the following document {obj}. Here's the exception:\n\n{e}"))
    os._exit(os.EX_OK)

def nf_in_object(nested_field_name, obj):
  for name_part in nested_field_name:
    if name_part not in obj: return False
    obj = obj[name_part]
  return obj

def clean_object(obj, nested_fields):
  for nf in nested_fields:
    obj.pop(nf[0], None)
  return obj

def denest(fields, obj):
  nested_fields = [field.split('.') for field in fields if '.' in field]
  for nested_field in nested_fields:
    nf_value = nf_in_object(nested_field, obj)
    if nf_value: obj['.'.join(nested_field)] = nf_value
  obj = clean_object(obj, nested_fields)
  return obj

def final_pw(args, log):
  pw = args['password'] if args['password'] != None else os.environ[args['secret_password']] if args['secret_password'] != None and args['secret_password'] in os.environ else ''
  if pw == '' and not args['silent_mode']:
    pw = (getpass.getpass("Enter your es instance password. If not needed, press ENTER:  ")).strip()
  if args['user'] == '' and pw != '': log.warning(wrap_orange('You set a password but not a user. You either forgot to set the user or you set a useless password. If something goes wrong with the authentication, this warning might ring a bell :)'))
  return pw

def add_timezone(date_string, timezone):
  try:
    return dateparser.parse(date_string).replace(tzinfo=timezone).isoformat()
  except:
    sys.exit(wrap_red(f"The date you set ({date_string}) (either --starting_date or --ending_date) is not in the valid iso8601 format (YYYY-MM-ddTHH:mm:ss) and the dateparser raised an exception. Please use the standard iso8601 format"))

def remove_duplicates(args, df):
  log = args['log']
  try: 
    if args['remove_duplicates']:
      log.info(wrap_blue('Removing possible duplicates from the dataframe'))
      df.drop_duplicates(subset=args['fields_to_export'], inplace=True)
    else:
      df.drop_duplicates(subset=['_id', *args['fields_to_export']], inplace=True)
    return df
  except Exception as e:
    sys.exit(wrap_red(f"Something went wrong when removing duplicates (set by user or the possible duplicates due to multiprocessing). The partial csv files won't be deleted. Here's the exception: \n\n{e}"))

def clean_dictionary(dictionaries, main_keys, dict_type, filename, log):
  dictionary = dictionaries[dict_type]
  if dict_type == '--rename_aggregations':
    aggregation_dict = dictionaries['--aggregation_types']
    for agg_key in aggregation_dict.keys():
      if agg_key not in dictionary.keys(): dictionary[agg_key] = f"estocsv__{agg_key}__{aggregation_dict[agg_key]}"
  for key in list(dictionary):
    if key not in main_keys: 
      log.warning(wrap_orange(f"\"{key}\" you set in your {dict_type} is not present among the columns of the raw data file \"{filename}\". It'll be removed from the aggregated column"))
      dictionary.pop(key, None)
  return dictionary

def clean_list(list_to_clean, main_keys):
  return [element for element in list_to_clean if element in main_keys]

def rename_df_columns(df, new_columns):
  return df.rename(columns=new_columns)

def aggregate_fields(filename, args, log, df_ready=None):
  try:
    df = pd.read_csv(filename, sep=args['csv_separator'], decimal=args['decimal_separator']) if df_ready is None else df_ready
    columns_to_keep = []
    aggregation_types = clean_dictionary({"--aggregation_types": args['aggregation_types']}, df.columns, '--aggregation_types', filename, log)
    aggregation_names = clean_dictionary({"--aggregation_types": args['aggregation_types'], "--rename_aggregations": args['rename_aggregations']}, df.columns, '--rename_aggregations', filename, log)
    boolean_occurrences = clean_list(args['count_boolean_occurrences'].split(','), df.columns)
    aggregation_fields = [ column for column in list(df.columns) if column in args['aggregation_fields'].split(',')]

    columns_to_keep = list(set(aggregation_fields + list(aggregation_types.keys()) + list(aggregation_names.keys()) + boolean_occurrences))

    for boolean in boolean_occurrences:
      df[boolean] = np.where(df[boolean] == True, 1, 0)
    
    agg_time_field = args['aggregation_time_field']
    if agg_time_field is not None and agg_time_field in df.columns:
      agg_timespan = args['aggregation_time_span']
      timezone = args['timezone'] if args['timezone'] is not None else tz.tzlocal()
      columns_to_keep = list(set(columns_to_keep + [agg_time_field]))
      df[agg_time_field] = pd.to_datetime(df[agg_time_field])
      df[agg_time_field] = df[agg_time_field].dt.tz_convert(timezone)
      df = df.set_index(agg_time_field)
      aggregation_fields.insert(0, pd.Grouper(level=agg_time_field, freq=f"{agg_timespan}d"))
      
    for column in df.columns:
      if column not in columns_to_keep: df.drop(columns=[column], inplace=True)
    
    aggregated_df = df.groupby(aggregation_fields).agg(aggregation_types)
    aggregated_df.rename(columns=aggregation_names, inplace=True)
    aggregated_df = aggregated_df.reset_index()

    if agg_time_field is not None and agg_time_field in aggregated_df.columns:
      aggregated_df[agg_time_field] = aggregated_df[agg_time_field].dt.date

    return aggregated_df
  except Exception as e:
    sys.exit(wrap_red(f"Something went wrong when trying to aggregate the raw documents. Here's the exception: \n\n{e}"))


from concurrent.futures import ProcessPoolExecutor
import multiprocessing, urllib3
from elasticsearch_tocsv.helpers.csv_handlers import *
from elasticsearch_tocsv.helpers.utility_functions import *
from elasticsearch_tocsv.helpers.interval_builders import *
from elasticsearch_tocsv.helpers.connection_tools import *
from elasticsearch_tocsv.helpers.arguments_checkers import *
from elasticsearch_tocsv.helpers.color_wrappers import *
from elasticsearch_tocsv.utils.TqdmLoggingHandler import TqdmLoggingHandler
from elasticsearch_tocsv.utils.CustomLogger import CustomLogger

def main():
  log = CustomLogger(__name__).logger
  args = fetch_arguments()

  if args['silent_mode']: log.warning(wrap_orange(f"You are running the tool in -sm/--silent_mode. Thus warning messages will be printed (e.g. SSL warnings, files overwriting warnings) but no confirmation will be asked.\n"))

  if not args['cert_verification']:
    urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

  args = check_arguments_conflicts(args, log)

  if args['aggregate_only'] == False:
    check_csv_already_written(args['export_path'], args['silent_mode'], log, 'query')

    log.info(wrap_green(bold("################### EXTRACTING DATA ###################\n")))

    if args['enable_multiprocessing']:
      log.info(wrap_blue('Connection to ES host established -- Multiprocessing enabled\n'))
      processes_to_use = min(multiprocessing.cpu_count(), args['process_number']) if args['process_number'] != None else multiprocessing.cpu_count()
      
      # Split total time range in equals time intervals so to let each process work on a partial set of data and store the resulting list as a sublist of the lists_to_join list
      if args['load_balance_interval'] == None:
        log.info(wrap_blue("No --load_balance_interval has been set. Intervals will be created on a time basis\n"))
        processes_intervals = make_time_intervals(args, processes_to_use, args['starting_date'], args['ending_date'])
      else:
        log.info(wrap_blue("Having set the --load_balance_interval option, intervals will be created on a load basis as far as possible according to the -lbi set\n"))
        processes_intervals = make_intervals_by_load(args, processes_to_use, args['starting_date'], args['ending_date'])

      # Check if to create aggregated csv file
      log.info(aggregation_log(args['aggregated_export_path'], args['aggregation_fields'], args['aggregation_types']))

      # Build the list of arguments to pass to the function each process will run
      process_function_arguments = [[args for i in range(processes_to_use)], *processes_intervals, [i for i in range(processes_to_use)]]
      
      processes_done = [None for i in range(processes_to_use)]

      # Create and start processes_to_use number of processes
      with ProcessPoolExecutor(processes_to_use) as executor:
        processes_done = executor.map(fetch_es_data, *process_function_arguments)

      while list(set(processes_done)) != [True]:
        continue
    else:
      log.info(wrap_blue('Connection to ES host established -- Single process run\n'))

      # Check if to create aggregated csv file
      log.info(aggregation_log(args['aggregated_export_path'], args['aggregation_fields'], args['aggregation_types']))
      
      fetch_es_data(args, args['starting_date'], args['ending_date'])

    # Joining partial csv files previously created into a single one
    log.info(wrap_blue('Joining partial csv files\n'))
    final_df = join_partial_csvs(args['export_path'][:-4], args['csv_separator'], args['decimal_separator'])
    
    # Remove duplicates
    final_df = remove_duplicates(args, final_df)
    final_df_newcolumns = rename_df_columns(final_df, args['rename_fields'])
    fields_to_export = [args['rename_fields'][field] if field in args['rename_fields'].keys() else field for field in args['fields_to_export']]

    # Write the CSV from the dataframe
    log.info(wrap_blue(f'Creating the csv at the following export path --> {args["export_path"]}.\n'))
    write_csv(args['export_path'], fields_to_export, args['csv_separator'], args['decimal_separator'], args['decimal_rounding'], "Something went wrong when trying to write the final csv after the merge of the partial csv files. The partial csv files won't be deleted.", df=final_df_newcolumns)

    # Delete partial csvs
    if not args['keep_partials']: 
      log.info(wrap_blue('Deleting partial csv files\n'))
      delete_partial_csvs(args['export_path'][:-4])

  # Create aggregation file if needed
  if args['aggregation_fields'] != None:
    check_csv_already_written(args['aggregated_export_path'], args['silent_mode'], log, 'aggregation')
    if args['aggregate_only']: 
      log.info(wrap_green(bold("################# AGGREGATING RAW FILE ################\n")))
    else:
      args['df_ready'] = final_df
    log.info(wrap_blue('Creating the csv file with aggregated data\n'))
    if not args['aggregation_fields_checked']: check_valid_agg_related_fields(args, log)
    file_to_aggregate = args['raw_input_file'] if args['aggregate_only'] else args['export_path']
    aggregated_df = aggregate_fields(file_to_aggregate, args, log, args['df_ready'])
    aggregated_df = rename_df_columns(aggregated_df, args['rename_fields'])
    write_csv(args['aggregated_export_path'], None,  args['csv_separator'], args['decimal_separator'], args['decimal_rounding'], f"Something when wrong when trying to write the aggregated dataframe to {args['aggregated_export_path']}. If you exported the raw data from Elasticsearch, the raw csv file will be available at \"{args['export_path']}\". Please use it with the --aggregate_only flag and a proper --aggregation_types and --aggregation_fields configuration to generate your aggregated file.", df=aggregated_df, index=False)
    log.info(wrap_blue(f"Aggregated file has been created at path \"{args['aggregated_export_path']}\"\n"))

  log.info(wrap_green(bold("################# EXITING SUCCESSFULLY ################\n")))

if __name__ == "__main__":
  main()




import os, sys, glob
import pandas as pd
from .color_wrappers import *

def check_csv_already_written(filename, silent_mode, log, operation_type):
  if os.path.exists(filename):
    overwrite_file = 'y' if silent_mode else ''
    log.warning(wrap_orange(f"There is already a csv file at the given path ({filename}) that will be overwritten."))
    while overwrite_file not in ['y', 'n']:
      overwrite_file = input(wrap_orange(f"\nDo you want to overwrite it? (y/n)")).lower().strip()
    if overwrite_file == 'n':
      sys.exit(wrap_red(wrap_blue(f"\nExiting script not to overwrite the file. No {operation_type} has been run.")))
    elif overwrite_file != 'y':
      print("Sorry, I can't understand the answer. Please answer with 'y' or 'n'")
  return filename

def write_csv(export_path, fields_to_export, separator, decimal_separator, decimal_rounding, exception_message='', df=None, list_to_convert=False, index=False, new_fields={}):
  df = pd.DataFrame(list_to_convert, columns=fields_to_export) if list_to_convert else df
  if df is not None: 
    df = df.round(decimal_rounding)
    try:
      if fields_to_export is None:
        df.to_csv(export_path, index=index, sep=separator, decimal=decimal_separator)
      else:
        df.to_csv(export_path, index=index, columns=fields_to_export, sep=separator, decimal=decimal_separator)
    except Exception as e:
      sys.exit(wrap_red(f"{exception_message} Here's the exception: \n\n{e}"))

def join_partial_csvs(filename, separator, decimal_separator):
  try: 
    list_of_csvs = sorted([csv for csv in glob.glob(f"{filename}_process*.csv")])
    final_df = pd.concat([pd.read_csv(csv, sep=separator, decimal=decimal_separator) for csv in list_of_csvs])
    return final_df
  except Exception as e:
    sys.exit(wrap_red(f"Something went wrong when trying to merge partial csv files previously created. The partial csv files won't be deleted. Here's the exception: \n\n{e}"))

def delete_partial_csvs(basic_filename):
  try:
    list_of_csvs = sorted([csv for csv in glob.glob(f"{basic_filename}_process*.csv")])
    for csv in list_of_csvs:
      os.remove(csv)
    return True
  except Exception as e:
    sys.exit(wrap_red(f"Something went wrong when trying to delete partial csv files previously created. Here's the exception: \n\n{e}"))